"""Iterator classes for notional."""

import logging
from abc import ABC, abstractmethod

CONTENT_PAGE_SIZE = 100

log = logging.getLogger(__name__)


class ContentIterator(ABC):
    """Base class to handle pagination over content from the Notion API."""

    def __init__(self):
        self.page = None
        self.index = -1
        self.pagenum = 0
        self.log = log.getChild("ContentIterator")

    def __iter__(self):
        """Initialize the iterator."""
        return self

    def __next__(self):
        """Return the next item from the result set or raise StopIteration."""
        # load a new page if needed
        if self.page is None or self.index >= len(self.page):
            self.index = 0
            self.page = self.load_next_page()
            self.pagenum += 1

        # if we have run out of results...
        if self.page is None or len(self.page) == 0:
            raise StopIteration

        # pull the next item from the current page
        item = self.page[self.index]

        # setup for the next call
        self.index += 1

        return item

    @property
    def page_number(self):
        """Return the current page number of results in this iterator."""
        return self.pagenum

    @abstractmethod
    def load_next_page(self):
        # noqa
        pass


class ResultSetIterator(ContentIterator, ABC):
    """Base class for iterating over result sets (using a cursor)."""

    def __init__(self):
        """Initialize the iterator."""
        super().__init__()
        self.cursor = None
        self.log = log.getChild("ResultSetIterator")

    def load_next_page(self):
        """Return the next page of content."""
        if self.cursor is False:
            return None

        params = {"page_size": CONTENT_PAGE_SIZE}

        if self.cursor:
            params["start_cursor"] = self.cursor

        self.log.debug("loading next page - start cursor: %s", self.cursor)

        # TODO error checking on result
        result = self.get_page(params)

        if result["has_more"]:
            self.cursor = result["next_cursor"]
        else:
            self.cursor = False

        results = result["results"]

        self.log.debug("loaded %d results; next cursor: %s", len(results), self.cursor)

        return result["results"]

    @property
    def last_page(self):
        """Return true if this is the last page of results."""
        if self.cursor is None:
            raise ValueError("iterator has not been initialized")

        return self.cursor is False

    @abstractmethod
    def get_page(self, params):
        # noqa
        pass


class EndpointIterator(ResultSetIterator):
    """Base class for iterating over results from an API endpoint."""

    def __init__(self, endpoint, **params):
        super().__init__()
        self.endpoint = endpoint
        self.params = params
        self.log = log.getChild("EndpointIterator")

    def get_page(self, params):
        """Return the next page with given parameters."""
        params.update(self.params)
        return self.endpoint(**params)



"""Wrapper for Notion API data types.

Similar to other records, these object provide access to the primitive data structure
used in the Notion API as well as higher-level methods.
"""

import logging
from datetime import date, datetime
from typing import Dict, List, Optional, Union
from uuid import UUID

from .core import DataObject, NestedObject, TypedObject
from .text import Color, FullColor, plain_text
from .user import User

log = logging.getLogger(__name__)

# TODO fix the Mention classes... they are not consistent with the rest of this library


class Annotations(DataObject):
    """Style information for RichTextObject's."""

    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    color: FullColor = None

    @property
    def is_plain(self):
        if self.bold:
            return False
        if self.italic:
            return False
        if self.strikethrough:
            return False
        if self.underline:
            return False
        if self.code:
            return False
        if self.color is not None:
            return False
        return True


class LinkObject(DataObject):
    """Reference a URL."""

    type: str = "url"
    url: str = None


class EmojiObject(TypedObject, type="emoji"):
    """A Notion emoji object."""

    emoji: str = None


class FileObject(TypedObject):
    """A Notion file object."""


class File(FileObject, type="file"):
    """A Notion file reference."""

    class NestedData(NestedObject):
        url: str = None
        expiry_time: datetime = None

    file: NestedData = None


class ExternalFile(FileObject, type="external"):
    """An external file reference."""

    class NestedData(NestedObject):
        url: str = None

    external: NestedData = None
    name: Optional[str] = None


class RichTextObject(TypedObject):
    """Base class for Notion rich text elements."""

    plain_text: str
    href: str = None
    annotations: Annotations = None

    def __str__(self):
        """Return a string representation of this object."""

        if self.href is None:
            text = self.plain_text
        else:
            text = f"[{self.plain_text}]({self.href})"

        # TODO add markdown for annotations
        # e.g. something like: text = markup(text, self.annotations)

        return text


class TextObject(RichTextObject, type="text"):
    """Notion text element."""

    class NestedData(NestedObject):
        content: str
        link: LinkObject = None

    text: NestedData = None

    @classmethod
    def from_value(cls, string):
        """Return a TextObject from the native string."""

        # TODO support markdown in the text string

        text = cls.NestedData(content=string)
        return cls(plain_text=string, text=text)


class MentionPageRef(DataObject):

    id: UUID


class MentionObject(RichTextObject, type="mention"):
    """Notion mention element."""

    class NestedData(NestedObject, TypedObject):
        pass

    class MentionUser(NestedData):
        user: User

    class MentionPage(NestedData):
        page: MentionPageRef

    class MentionDatabase(NestedData):
        database: MentionPageRef

    mention: NestedData = None


class EquationObject(RichTextObject, type="equation"):
    """Notion equation element."""

    class NestedData(NestedObject):
        expression: str

    equation: NestedData = None

    def __str__(self):
        """Return a string representation of this object."""

        if self.equation is None:
            return None

        return self.equation.expression


class NativeTypeMixin(object):
    """Mixin class for properties that can be represented as native Python types."""

    def __str__(self):
        """Return a string representation of this object."""

        return str(self.Value or "")

    def __eq__(self, other):
        """Determine if this property is equal to the given object."""

        return self.Value == other

    def __ne__(self, other):
        """Determine if this property is not equal to the given object."""

        return self.Value != other

    @property
    def Value(self):
        """Get the current value of this property as a native Python type."""

        cls = self.__class__

        # check to see if the object has a field with the type-name
        # (this is assigned by TypedObject during subclass creation)
        if hasattr(cls, "type") and hasattr(self, cls.type):
            return getattr(self, cls.type)

        raise NotImplementedError()

    @classmethod
    def from_value(cls, value):
        """Build the property value from the native Python value."""

        # use type-name field to instantiate the class when possible
        if hasattr(cls, "type"):
            return cls(**{cls.type: value})

        raise NotImplementedError()


class PropertyValue(TypedObject):
    """Base class for Notion property values."""

    id: str = None


class Title(NativeTypeMixin, PropertyValue, type="title"):
    """Notion title type."""

    title: List[RichTextObject] = []

    def __len__(self):
        """Return the number of object in the Title object."""

        return len(self.title)

    @property
    def Value(self):
        """Return the plain text from this Title."""

        if self.title is None:
            return None

        return plain_text(*self.title)

    @classmethod
    def from_value(cls, value):
        text = TextObject.from_value(value)
        return cls(title=[text])


class RichText(NativeTypeMixin, PropertyValue, type="rich_text"):
    """Notion rich text type."""

    rich_text: List[RichTextObject] = []

    def __len__(self):
        """Return the number of object in the RichText object."""
        return len(self.rich_text)

    @property
    def Value(self):
        """Return the plain text from this RichText."""

        if self.rich_text is None:
            return None

        return plain_text(*self.rich_text)

    @classmethod
    def from_value(cls, value):
        text = TextObject.from_value(value)
        return cls(rich_text=[text])


class Number(NativeTypeMixin, PropertyValue, type="number"):
    """Simple number type."""

    number: Union[float, int] = None

    def __iadd__(self, other):
        """Add the given value to this Number."""

        self.number += other
        return self

    def __isub__(self, other):
        """Subtract the given value from this Number."""

        self.number -= other
        return self


class Checkbox(NativeTypeMixin, PropertyValue, type="checkbox"):
    """Simple checkbox type; represented as a boolean."""

    checkbox: bool = None


class DateRange(DataObject):
    """A Notion date range, with an optional end date."""

    start: Union[date, datetime]
    end: Optional[Union[date, datetime]] = None


class Date(PropertyValue, type="date"):
    """Notion complex date type - may include timestamp and/or be a date range."""

    date: DateRange = None

    def __str__(self):
        """Return a string representation of this object."""

        if self.date is None:
            return None

        if self.date.end is None:
            return f"{self.date.start}"

        return f"{self.date.start} :: {self.date.end}"

    def __contains__(self, other):
        """Determines if the given date is in the range (inclusive) of this Date.

        Raises ValueError if the Date object is not a range - e.g. has no end date.
        """

        if not self.IsRange:
            raise ValueError("This date is not a range")

        return self.Start <= other <= self.End

    @property
    def IsRange(self):
        """Indicates if this object represents a date range (versus a single date)."""

        if self.date is None:
            return False

        return self.date.end is not None

    @property
    def Start(self):
        return None if self.date is None else self.date.start

    @property
    def End(self):
        return None if self.date is None else self.date.end

    @classmethod
    def from_value(cls, value):
        """Create a new Date from the native value."""
        inner = DateRange(start=value)
        return cls(date=inner)


class SelectValue(DataObject):
    """Values for select & multi-select properties."""

    name: str
    id: UUID = None
    color: Color = None


class SelectOne(NativeTypeMixin, PropertyValue, type="select"):
    """Notion select type."""

    select: SelectValue = None

    def __str__(self):
        """Return a string representation of this object."""

        return self.select.name or self.select.id or ""

    def __eq__(self, other):
        """Determine if this property is equal to the given object.

        To avoid confusion, this method only compares Select options by name, not ID.
        """

        if self.select is None:
            return other == None

        return other == self.select.name

    @property
    def Value(self):
        if self.select is None:
            return None

        return self.select.name or self.select.id

    @classmethod
    def from_value(cls, value):
        return cls(select=SelectValue(name=value))


class MultiSelect(NativeTypeMixin, PropertyValue, type="multi_select"):
    """Notion multi-select type."""

    multi_select: List[SelectValue] = []

    def __str__(self):
        """Return a string representation of this object."""
        return ", ".join(self.Values)

    def __iadd__(self, other):
        """Add the given option to this MultiSelect."""

        if other in self:
            raise ValueError(f"Item already selected: {other}")

        opt = SelectValue(name=other)
        self.multi_select.append(opt)

        return self

    def __isub__(self, other):
        """Remove the given value from this MultiSelect."""

        self.multi_select = [opt for opt in self.multi_select if opt.name != other]

        return self

    def __contains__(self, name):
        """Determines if the given name is in this MultiSelect.

        To avoid confusion, only names are considered for comparison, not ID's.
        """

        for opt in self.multi_select:
            if opt.name == name:
                return True

        return False

    @property
    def Value(self):
        if self.multi_select is None:
            return None

        values = list()

        for select in self.multi_select:
            if select.name is not None:
                values.append(select.name)
            elif select.id is not None:
                values.append(select.id)

        return values

    @classmethod
    def from_value(cls, value):

        values = [SelectValue(name=item) for item in value if item is not None]

        return cls(multi_select=values)


class People(PropertyValue, type="people"):
    """Notion people type."""

    people: List[User] = []

    def __iter__(self):
        """Iterates over the User's in this property."""

        if self.people is None:
            return None

        return iter(self.people)

    def __contains__(self, other):
        """Determines if the given User or name is in this People.

        To avoid confusion, only names are considered for comparison (not ID's).
        """

        for user in self.people:
            if user == other:
                return True
            elif user.name == other:
                return True

        return False


class URL(NativeTypeMixin, PropertyValue, type="url"):
    """Notion URL type."""

    url: str = None


class Email(NativeTypeMixin, PropertyValue, type="email"):
    """Notion email type."""

    email: str = None


class PhoneNumber(NativeTypeMixin, PropertyValue, type="phone_number"):
    """Notion phone type."""

    phone_number: str = None


class Files(PropertyValue, type="files"):
    """Notion files type."""

    files: List[FileObject] = []

    def __contains__(self, other):
        """Determines if the given FileObject or name is in the file list.

        To avoid confusion, only names are considered for comparison.
        """

        if self.files is None:
            return False

        for file in self.files:
            if file == other:
                return True
            elif file.name == other:
                return True

        return False


class FormulaResult(TypedObject):
    """A Notion formula result.

    This object contains the result of the expression in the database properties.
    """

    @property
    def Result(self):
        """Return the result of this FormulaResult."""

        raise NotImplemented("Result unavailable")


class StringFormula(FormulaResult, type="string"):
    """A Notion string formula result."""

    string: str = None

    def __str__(self):
        return self.string or ""

    @property
    def Result(self):
        """Return the result of this StringFormula."""

        return self.string


class NumberFormula(FormulaResult, type="number"):
    """A Notion number formula result."""

    number: Union[int, float] = None

    def __str__(self):
        return str(self.number or "")

    @property
    def Result(self):
        """Return the result of this NumberFormula."""

        return self.number


class DateFormula(FormulaResult, type="date"):
    """A Notion date formula result."""

    date: DateRange = None

    def __str__(self):
        return str(self.date or "")

    @property
    def Result(self):
        """Return the result of this DateFormula."""

        return self.date


class Formula(PropertyValue, type="formula"):
    """A Notion formula property value."""

    formula: FormulaResult = None

    def __str__(self):
        return str(self.Result or "")

    @property
    def Result(self):
        """Return the result of this Formula in its native type."""

        if self.formula is None:
            return None

        return self.formula.Result


class Relation(PropertyValue, type="relation"):
    """A Notion relation property value."""

    relation: List[str] = []


class CreatedTime(NativeTypeMixin, PropertyValue, type="created_time"):
    """A Notion created-time property value."""

    created_time: datetime = None


class CreatedBy(PropertyValue, type="created_by"):
    """A Notion created-by property value."""

    created_by: User = None


class LastEditedTime(NativeTypeMixin, PropertyValue, type="last_edited_time"):
    """A Notion last-edited-time property value."""

    last_edited_time: datetime = None


class LastEditedBy(PropertyValue, type="last_edited_by"):
    """A Notion last-edited-by property value."""

    last_edited_by: User = None



"""Wrapper for Notion User objects."""

import logging
from uuid import UUID

from .core import NestedObject, TypedObject

log = logging.getLogger(__name__)


class User(TypedObject):
    """Represents a User in Notion."""

    object: str = "user"
    id: UUID = None
    name: str = None
    avatar_url: str = None


class Person(User, type="person"):
    """Represents a Person in Notion."""

    class NestedPerson(NestedObject):
        email: str

    person: NestedPerson = None

    def __getitem__(self, key):
        return self.person[key]


class Bot(User, type="bot"):
    """Represents a Bot in Notion."""

    class NestedBot(NestedObject):
        pass

    bot: NestedBot = None

    def __getitem__(self, key):
        return self.bot[key]



"""Wrapper for Notion API data types.

We handle databases and pages somewhat differently than other blocks since they
represent top-level containers for other blocks.

These objects provide both access to the primitive data structure returned by the API
as well as higher-level access methods.  In general, attributes in lower case represent
the primitive data structure, where capitalized attributes provide higher-level access.
"""

import logging
from datetime import datetime
from typing import Dict, List, Optional, Union
from uuid import UUID

from .core import NamedObject, TypedObject
from .schema import Schema
from .text import plain_text
from .types import EmojiObject, FileObject, PropertyValue, RichTextObject

log = logging.getLogger(__name__)


class ParentRef(TypedObject):
    """Reference another block."""

    # TODO method / property to resolve the reference?


class DatabaseParent(ParentRef, type="database_id"):
    """Reference a database."""

    database_id: UUID


class PageParent(ParentRef, type="page_id"):
    """Reference a page."""

    page_id: UUID


class WorkspaceParent(ParentRef, type="workspace"):
    """Reference the workspace."""

    workspace: bool = True


class Record(NamedObject):
    """The base type for Notion API records."""

    id: UUID = None
    created_time: datetime = None
    last_edited_time: datetime = None
    has_children: bool = False


class Database(Record, object="database"):
    """A database record type."""

    title: List[RichTextObject] = None
    parent: ParentRef = None
    icon: Optional[Union[FileObject, EmojiObject]] = None
    cover: Optional[FileObject] = None
    properties: Schema = {}

    @property
    def Title(self):
        if self.title is None:
            return None

        return plain_text(*self.title)


class Page(Record, object="page"):
    """A standard Notion page object."""

    archived: bool = False
    url: str = None
    parent: ParentRef = None
    icon: Optional[Union[FileObject, EmojiObject]] = None
    cover: Optional[FileObject] = None
    properties: Dict[str, PropertyValue] = {}

    def __getitem__(self, name):
        """Indexer for the given property name.

        :param name: the name of the property to get
        """

        log.debug("get property :: {%s} [%s]", self.id, name)

        if self.properties is None:
            raise AttributeError("No properties in Page")

        prop = self.properties.get(name)

        if prop is None:
            raise AttributeError(f"No such property: {name}")

        return prop

    def __setitem__(self, name, prop):
        """Set the object data for the given property.

        :param name: the name of the property to set
        :param prop: the PropertyValue for the named property
        """

        log.debug("set property :: {%s} [%s] => %s", self.id, name, prop)

        self.properties[name] = prop

    @property
    def Title(self):
        if self.properties is None:
            return None

        for prop in self.properties.values():
            if prop.id == "title":
                return prop.Value

        return None



"""Wrapper for Notion API blocks.

Blocks are specifc records that hold content.

Similar to other records, these object provide access to the primitive data structure
used in the Notion API as well as higher-level methods.
"""

import logging
from typing import Any, Dict, List, Optional

from .core import NestedObject, TypedObject
from .records import Record
from .types import FileObject, RichTextObject, TextObject

log = logging.getLogger(__name__)


class Block(Record, TypedObject, object="block"):
    """A standard block object in Notion."""


class UnsupportedBlock(Block, type="unsupported"):
    """A placeholder for unsupported blocks in the API."""

    unsupported: Optional[Dict[str, Any]] = None


class TextBlock(Block):
    """A standard text block object in Notion."""

    @classmethod
    def from_text(cls, text):
        text = TextObject.from_value(text)

        if not hasattr(cls, "type") or cls.type is None:
            raise TypeError(f"class type is not defined: {cls}")

        # text types have a nested object with 'type' name and a 'text' child
        # here, we use the constructor to build out the nested object...

        return cls(**{cls.type: {"text": [text]}})


class Paragraph(TextBlock, type="paragraph"):
    """A paragraph block in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None
        children: List[Block] = None

    paragraph: NestedData = None


class Heading1(TextBlock, type="heading_1"):
    """A heading_1 block in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None

    heading_1: NestedData = None


class Heading2(TextBlock, type="heading_2"):
    """A heading_2 block in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None

    heading_2: NestedData = None


class Heading3(TextBlock, type="heading_3"):
    """A heading_3 block in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None

    heading_3: NestedData = None


class BulletedListItem(TextBlock, type="bulleted_list_item"):
    """A bulleted list item in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None
        children: List[Block] = None

    bulleted_list_item: NestedData = None


class NumberedListItem(TextBlock, type="numbered_list_item"):
    """A numbered list item in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None
        children: List[Block] = None

    numbered_list_item: NestedData = None


class ToDo(TextBlock, type="to_do"):
    """A todo list item in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None
        checked: bool = False
        children: List[Block] = None

    to_do: NestedData = None


class Toggle(TextBlock, type="toggle"):
    """A toggle list item in Notion."""

    class NestedData(NestedObject):
        text: List[RichTextObject] = None
        children: List[Block] = None

    toggle: NestedData = None


class Embed(Block, type="embed"):
    """An embed block in Notion."""

    class NestedData(NestedObject):
        url: str

    embed: NestedData = None

    @classmethod
    def from_url(cls, url):
        obj = NestedData(url=url)
        return Embed(embed=obj)


class Bookmark(Block, type="bookmark"):
    """A bookmark block in Notion."""

    class NestedData(NestedObject):
        url: str

    bookmark: NestedData = None

    @classmethod
    def from_url(cls, url):
        obj = NestedData(url=url)
        return Bookmark(bookmark=obj)


class File(Block, type="file"):
    """A file block in Notion."""

    file: FileObject = None


class Image(Block, type="image"):
    """An image block in Notion."""

    image: FileObject = None


class Video(Block, type="video"):
    """A video block in Notion."""

    video: FileObject = None


class PDF(Block, type="pdf"):
    """A pdf block in Notion."""

    pdf: FileObject = None


class ChildPage(Block, type="child_page"):
    """A child page block in Notion."""

    class NestedData(NestedObject):
        title: str

    child_page: NestedData = None



"""Base classes for working with the Notion API."""

import inspect
import json
import logging
from datetime import date, datetime
from uuid import UUID

from pydantic import BaseModel, validator
from pydantic.main import validate_model

log = logging.getLogger(__name__)


def make_api_safe(data):
    """Recursively convert the given data to an API-safe form.

    This is mostly to handle data types that will not directly serialize to JSON.
    """

    if isinstance(data, date) or isinstance(data, datetime):
        return data.isoformat()

    elif isinstance(data, UUID):
        return data.hex

    elif isinstance(data, dict):
        return {name: make_api_safe(value) for name, value in data.items()}

    elif isinstance(data, list):
        return [make_api_safe(value) for value in data]

    elif isinstance(data, tuple):
        return [make_api_safe(value) for value in data]

    return data


class DataObject(BaseModel):
    """The base for all API objects."""

    # See https://github.com/samuelcolvin/pydantic/issues/1577
    def __setattr__(self, name, value):
        try:
            super().__setattr__(name, value)
        except ValueError as e:
            setters = inspect.getmembers(
                self.__class__,
                predicate=lambda x: isinstance(x, property) and x.fset is not None,
            )
            for setter_name, func in setters:
                if setter_name == name:
                    object.__setattr__(self, name, value)
                    break
            else:
                raise e

    def refresh(__pydantic_self__, **data):
        """Refresh the internal attributes with new data."""

        # log.debug("validating object data -- %s", data)

        values, fields, error = validate_model(__pydantic_self__.__class__, data)

        if error:
            raise error

        log.debug("refreshing object values -- %s", values)

        object.__setattr__(__pydantic_self__, "__dict__", values)

    def to_api(self):
        """Convert to a suitable representation for the Notion API."""

        # the API doesn't like "undefined" values...

        data = self.dict(exclude_none=True)

        # we need to convert "special" types to string forms to help the JSON encoder.
        # there are efforts underway in pydantic to make this easier, but for now...

        # XXX this method has promise, but converts the entire value to a single string
        # return {name: json.dumps(value, default=str) for name, value in data.items()}

        return make_api_safe(data)


class NamedObject(DataObject):
    """A Notion API object."""

    object: str = None

    def __init_subclass__(cls, object=None, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.object = object


class TypedObject(DataObject):
    """A type-referenced object.

    Many objects in the Notion API follow a generic->specific pattern with a 'type'
    parameter followed by additional data.  These objects must specify a `type`
    attribute to ensure that the correct object is created.
    """

    type: str = None

    # modified from the methods described in this discussion:
    # - https://github.com/samuelcolvin/pydantic/discussions/3091

    def __init_subclass__(cls, type=None, **kwargs):
        """Register the subtypes of the TypedObject subclass."""
        super().__init_subclass__(**kwargs)

        if type is not None:
            sub_type = type

            # also set the default for the class 'type' field
            cls.type = sub_type

        elif hasattr(cls, "__type__"):
            sub_type = getattr(cls, "__type__")

        else:
            sub_type = cls.__name__

        # initialize a _subtypes_ map for each direct child of TypedObject

        # this allows different class trees to have the same 'type' name
        # but point to a different object (e.g. the 'date' type may have
        # different implementations depending where it is used in the API)

        # also, due to the order in which typed classes are defined, once
        # the map is defined for a subclass of TypedObject, any further
        # descendants of that class will have the new map via inheritance

        if TypedObject in cls.__bases__ and not hasattr(cls, "_subtypes_"):
            cls._subtypes_ = dict()

        if sub_type in cls._subtypes_:
            raise ValueError(f"Duplicate subtype for class - {sub_type} :: {cls}")

        log.debug("registered new subtype: %s => %s", sub_type, cls)

        cls._subtypes_[sub_type] = cls

    @classmethod
    def __get_validators__(cls):
        yield cls._convert_to_real_type_

    @classmethod
    def parse_obj(cls, obj):
        return cls._convert_to_real_type_(obj)

    @classmethod
    def _convert_to_real_type_(cls, data):
        """Instantiate the correct object based on the 'type' field."""

        if isinstance(data, cls):
            return data

        if not isinstance(data, dict):
            raise ValueError("Invalid 'data' object")

        data_type = data.get("type")

        if data_type is None:
            raise ValueError("Missing 'type' in TypedObject")

        if not hasattr(cls, "_subtypes_"):
            raise TypeError(f"Invalid TypedObject: {cls} - missing _subtypes_")

        sub = cls._subtypes_.get(data_type)

        if sub is None:
            raise TypeError(f"Unsupport sub-type: {data_type}")

        log.debug(
            "initializing typed object %s :: %s => %s -- %s", cls, data_type, sub, data
        )

        return sub(**data)


class NestedObject(DataObject):
    """Represents an API object with nested data.

    These objects require a 'type' property and a matching property of the same
    name, which holds additional data.

    For example, this contains a nested 'text' object:

        data = {
            type: "text",
            ...
            text: {
                ...
            }
        }

    Currently, this is a convenience class for clarity - it does not provide additional
    functionality at this time.
    """

    # XXX can we somehow help provide pass-through access from the outer class?

    # some ideas...
    #   use __call__ of the outer class to access inner metadata object
    #   use __getitem__ of the outer class to access inner metadata properties



"""Utilities for working with Notion as an ORM."""

import logging

from .records import Page
from .types import NativeTypeMixin, PropertyValue, RichText

log = logging.getLogger(__name__)


class ConnectedPageBase(object):
    """Base class for "live" pages via the Notion API."""

    def __init__(self, **data):
        self.page = Page(**data)

    @property
    def id(self):
        """Return the ID of this page (if available)."""

        return None if self.page is None else self.page.id

    @property
    def children(self):
        """Return an iterator for all child blocks of this Page."""

        return EndpointIterator(
            endpoint=self._session.blocks.children.list, block_id=self.id
        )

    def append(self, *blocks):
        """Append the given blocks as children of this ConnectedPage."""

        data = [block.to_api() for block in blocks]

        self._orm_session_.blocks.children.append(block_id=self.page.id, children=data)

    def kwargs_to_props(self, **kwargs):
        """Converts the list of keyword args to a dict of properties."""

        props = dict()

        for name, prop in properties.items():
            if isinstance(prop, PropertyValue):
                props[name] = prop.to_api()

            elif isinstance(prop, dict):
                props[name] = prop

            elif hasattr(cls, name):
                field = getattr(cls, name)
                prop = field.from_value(prop)  # FIXME
                props[name] = prop.to_api()

            else:
                raise ValueError(f"Unsupported property: {name}")

        return props

    @classmethod
    def create(cls, **properties):
        """Creates a new instance of the ConnectedPage type."""

        log.debug(f"creating new {cls} :: {cls._orm_database_id_}")

        parent_id = {"database_id": cls._orm_database_id_}

        # TODO convert properties to a dict for create...
        # props = kwargs_to_props(properties)
        props = dict()

        data = cls._orm_session_.pages.create(parent=parent_id, properties=props)

        return cls(**data)

    @classmethod
    def parse_obj(cls, data):
        return cls(**data)


def Property(name, cls=RichText, default=None):
    """Define a property for a Notion Page object.

    :param name: the Notion table property name
    :param cls: the data type for the property (default = RichText)
    :param default: a default value when creating new objects
    """

    log.debug("creating new Property: %s", name)

    def fget(self):
        """Return the current value of the property as a python object."""

        if not isinstance(self, ConnectedPageBase):
            raise TypeError("Properties must be used in a ConnectedPage object")

        try:
            prop = self.page[name]

        except AttributeError:
            log.debug(f"property '{name}' does not exist in Page; returning default")
            return default

        if not isinstance(prop, cls):
            raise TypeError(f"Type mismatch: expected '{cls}' but found '{type(prop)}'")

        # convert native objects to expected types
        if isinstance(prop, NativeTypeMixin):
            return prop.Value

        return prop

    def fset(self, value):
        """Set the property to the given value."""

        if not isinstance(self, ConnectedPageBase):
            raise TypeError("Properties must be used in a ConnectedPage object")

        # TODO only set the value if it has changed from the existing
        log.debug(f"set {cls} [{name}] => {value} {type(value)}")

        # convert from native objects to expected types
        if isinstance(value, cls):
            prop = value
        elif issubclass(cls, NativeTypeMixin):
            prop = cls.from_value(value)
        else:
            raise ValueError(f"Value does not match expected type: {cls}")

        # commit the changes directly to Notion
        props = {name: prop.to_api()}
        data = self._orm_session_.pages.update(self.page.id, properties=props)

        # reset the internal page with the latest data
        self.page.refresh(**data)

    return property(fget, fset)


def connected_page(session, bind=ConnectedPageBase):
    """Returns a base class for "connected" pages through the Notion API."""

    if not issubclass(bind, ConnectedPageBase):
        raise ValueError("bind class must subclass ConnectedPageBase")

    class _ConnectedPage(bind):
        _orm_session_ = None
        _orm_database_id_ = None

        def __init_subclass__(cls, database, **kwargs):
            """Attach the ConnectedPage to the given database ID."""
            super().__init_subclass__(**kwargs)

            if database is None:
                raise ValueError("Missing 'database' for ConnectedPage: {cls}")

            if cls._orm_database_id_ is not None:
                raise TypeError("Object {cls} registered to: {database}")

            cls._orm_session_ = session
            cls._orm_database_id_ = database

            log.debug(f"registered connected page :: {cls} => {database}")

    return _ConnectedPage



"""Provide a simple ORM for Notion objects."""

import logging

from .session import Session
from .version import __version__

log = logging.getLogger(__name__)


def connect(**kwargs):
    """Connect to Notion using the provided integration token."""

    log.debug("connecting to Notion...")

    return Session(**kwargs)



"""Provides an interactive query builder for Notion databases."""

import logging

import notion_client

from .iterator import EndpointIterator
from .orm import ConnectedPageBase

log = logging.getLogger(__name__)

# TODO add support for start_cursor and page_size - the challenge is that we are using
# EndpointIterator for the results, which overrides those parameters for all results


class Query(object):
    """A query builder for the Notion API.

    :param session: an active session with the Notion SDK
    :param target: either a string with the database ID or an ORM class
    """

    def __init__(self, session, target):
        self.session = session
        self.target = target

        self._filter = list()
        self._sort = list()
        self._start = None
        self._limit = None

    def filter(self, *filters):
        """Add the given filter elements to the query."""

        # XXX see notes on sort()...

        self._filter.extend(filters)

        return self

    def sort(self, *sorts):
        """Add the given sort elements to the query."""

        # XXX should this support ORM properties also?
        # e.g. - query.sort(Task.Title, query.ASC)
        # but users won't always use ORM for queries...

        self._sort.extend(sorts)

        return self

    # def start_at(self, page_id):
    #     """Set the start cursor to a specific page ID."""
    #     self._start = page_id
    #     return self

    # def limit(self, page_size):
    #     """Limit the number of results to the given page size."""
    #     self._start = page_id
    #     return self

    def execute(self):
        """Execute the current query and return an iterator for the results."""

        params = {"endpoint": self.session.databases().query}

        cls = None

        if isinstance(self.target, str):
            params["database_id"] = self.target

        elif issubclass(self.target, ConnectedPageBase):
            cls = self.target

            if cls._orm_session_ != self.session:
                raise ValueError("ConnectedPage belongs to a different session")

            if cls._orm_database_id_ is None:
                raise ValueError("ConnectedPage has no database")

            params["database_id"] = cls._orm_database_id_

        else:
            raise ValueError("unsupported query target")

        if self._filter and len(self._filter) > 0:
            params["filter"] = self._filter

        if self._sort and len(self._sort) > 0:
            params["sorts"] = self._sort

        if self._start is not None:
            params["start_cursor"] = self._start

        if self._limit is not None:
            params["page_size"] = self._limit

        log.debug("executing query - %s", params)

        # FIXME in order to use start and limit, we need a different
        # mechanism for iterating on results
        items = EndpointIterator(**params)

        return ResultSet(session=self.session, src=items, cls=cls)

    def first(self):
        """Execute the current query and return the first result only."""

        try:
            return next(self.execute())
        except StopIteration:
            log.debug("iterator returned empty result set")

        return None


class ResultSet(object):
    """A result for a specific query."""

    def __init__(self, session, src, cls=None):
        self.session = session
        self.source = src
        self.cls = cls

    def __iter__(self):
        return self

    def __next__(self):
        item = next(self.source)

        if self.cls is not None:
            item = self.cls.parse_obj(item)

        return item



"""Objects representing a database schema."""

from typing import Any, Dict, List, Optional
from uuid import UUID

from .core import DataObject, NestedObject, TypedObject


class PropertyObject(TypedObject):
    """Base class for Notion property objects."""

    id: Optional[str] = None
    name: Optional[str] = None


Schema = Dict[str, PropertyObject]
"""A database schema, mapping property names to object configurations."""


class Title(PropertyObject, type="title"):
    """Defines the title configuration for a database property."""

    title: Any = {}


class RichText(PropertyObject, type="rich_text"):
    """Defines the rich text configuration for a database property."""

    rich_text: Any = {}


class Number(PropertyObject, type="number"):
    """Defines the number configuration for a database property."""

    class NestedData(NestedObject):
        format: str = "number"

    number: NestedData = {}

    @classmethod
    def format(cls, format):
        return cls(number=cls.NestedData(format=format))


class SelectOption(DataObject):
    """Options for select & multi-select objects."""

    name: str
    id: str = None
    color: str = None


class Select(PropertyObject, type="select"):
    """Defines the select configuration for a database property."""

    class NestedData(NestedObject):
        options: List[SelectOption] = []

    select: NestedData = None

    @classmethod
    def options(cls, *options):
        return cls(select=cls.NestedData(options=options))


class MultiSelect(PropertyObject, type="multi_select"):
    """Defines the multi-select configuration for a database property."""

    class NestedData(NestedObject):
        options: List[SelectOption] = []

    multi_select: NestedData = None


class Date(PropertyObject, type="date"):
    """Defines the date configuration for a database property."""

    date: Any = {}


class People(PropertyObject, type="people"):
    """Defines the people configuration for a database property."""

    people: Any = {}


class Files(PropertyObject, type="files"):
    """Defines the files configuration for a database property."""

    files: Any = {}


class Checkbox(PropertyObject, type="checkbox"):
    """Defines the checkbox configuration for a database property."""

    checkbox: Any = {}


class Email(PropertyObject, type="email"):
    """Defines the email configuration for a database property."""

    email: Any = {}


class URL(PropertyObject, type="url"):
    """Defines the URL configuration for a database property."""

    url: Any = {}


class PhoneNumber(PropertyObject, type="phone_number"):
    """Defines the phone number configuration for a database property."""

    phone_number: Any = {}


class Formula(PropertyObject, type="formula"):
    """Defines the formula configuration for a database property."""

    class NestedData(NestedObject):
        expression: str

    formula: NestedData = None


class Relation(PropertyObject, type="relation"):
    """Defines the relation configuration for a database property."""

    class NestedData(NestedObject):
        database_id: UUID
        synced_property_name: str = None
        synced_property_id: str = None

    relation: NestedData = None


class Rollup(PropertyObject, type="rollup"):
    """Defines the rollup configuration for a database property."""

    class NestedData(NestedObject):
        relation_property_name: str
        relation_property_id: str
        rollup_property_name: str
        rollup_property_id: str
        function: str

    rollup: NestedData = None


class CreatedTime(PropertyObject, type="created_time"):
    """Defines the created-time configuration for a database property."""

    created_time: Any = {}


class CreatedBy(PropertyObject, type="created_by"):
    """Defines the created-by configuration for a database property."""

    created_by: Any = {}


class LastEditedBy(PropertyObject, type="last_edited_by"):
    """Defines the last-edited-by configuration for a database property."""

    last_edited_by: Any = {}


class LastEditedTime(PropertyObject, type="last_edited_time"):
    """Defines the last-edited-time configuration for a database property."""

    last_edited_time: Any = {}



"""Handle session management with the Notion SDK."""

import logging

import notion_client

from .blocks import Block
from .iterator import EndpointIterator
from .query import Query, ResultSet
from .records import Database, Page, ParentRef
from .types import TextObject, Title
from .user import User

log = logging.getLogger(__name__)

# TODO add support for limits and filters in list() methods...


class Endpoint(object):
    """Notional wrapper for the API endpoints."""

    def __init__(self, session):
        self.session = session


class BlocksEndpoint(Endpoint):
    """Notional interface to the API 'blocks' endpoint."""

    class ChildrenEndpoint(Endpoint):
        """Notional interface to the API 'blocks/children' endpoint."""

        def __call__(self):
            return self.session.client.blocks.children

        # https://developers.notion.com/reference/patch-block-children
        def append(self, parent, *blocks):
            """Adds the given blocks as children of the specified parent.

            The parent info will be refreshed to the latest version from the server.
            """

            parent_id = parent.id

            children = [block.to_api() for block in blocks if block is not None]

            log.info("Appending %d blocks to %s...", len(children), parent_id)

            data = self().append(block_id=parent_id, children=children)

            parent.refresh(**data)

        # https://developers.notion.com/reference/get-block-children
        def list(self, parent):
            """Returns all Blocks contained by the specified parent."""

            blocks = EndpointIterator(endpoint=self().list, block_id=parent.id)

            log.info("Listing blocks for %s...", parent.id)

            return ResultSet(session=self, src=blocks, cls=Block)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.children = BlocksEndpoint.ChildrenEndpoint(*args, **kwargs)

    def __call__(self):
        return self.session.client.blocks

    # https://developers.notion.com/reference/retrieve-a-block
    def retrieve(self, block_id):
        """Returns the Block with the given ID."""

        log.info("Retrieving block :: %s", block_id)

        return Block.parse_obj(self().retrieve(block_id))

    # https://developers.notion.com/reference/update-a-block
    def update(self, block):
        """Update the block content on the server.

        The block info will be refreshed to the latest version from the server.
        """

        log.info("Updating block :: %s", block.id)

        data = self().update(block.id, **block.to_api())

        block.refresh(**data)


class DatabasesEndpoint(Endpoint):
    """Notional interface to the API 'databases' endpoint."""

    def __call__(self):
        return self.session.client.databases

    # https://developers.notion.com/reference/create-a-database
    def create(self, parent, schema, title=None):
        """Adds a database to the given Page parent."""

        parent_id = get_parent_id(parent)

        log.info("Creating database [%s] - %s", parent_id, title)

        props = {name: prop.to_api() for name, prop in schema.items()}

        if title is not None:
            title = TextObject.from_value(title)

        data = self().create(
            parent=parent_id,
            title=[title.to_api()],
            properties=props,
        )

        return Database.parse_obj(data)

    # https://developers.notion.com/reference/get-databases
    def list(self):
        """Returns an iterator for all Database objects in the integration scope."""

        log.info("Listing known databases...")

        databases = EndpointIterator(endpoint=self().list)
        return ResultSet(session=self, src=databases, cls=Database)

    # https://developers.notion.com/reference/retrieve-a-database
    def retrieve(self, database_id):
        """Returns the Database with the given ID."""

        log.info("Retrieving database :: %s", database_id)

        return Database.parse_obj(self().retrieve(database_id))

    # https://developers.notion.com/reference/update-a-database
    def update(self, database):
        """Updates the Database object on the server.

        The database info will be refreshed to the latest version from the server.
        """

        log.info("Updating database info :: ", database.id)

        data = self().update(database.id, **database.to_api())

        database.refresh(**data)

    # https://developers.notion.com/reference/post-database-query
    def query(self, target):
        """Initialized a new Query object with the target data class.

        :param target: either a string with the database ID or an ORM class
        """

        log.info("Initializing database query :: %s", target)

        return Query(self.session, target)


class PagesEndpoint(Endpoint):
    """Notional interface to the API 'pages' endpoint."""

    def __call__(self):
        return self.session.client.pages

    # https://developers.notion.com/reference/post-page
    def create(self, parent, title=None, properties=dict(), children=list()):
        """Adds a page to the given parent (Page or Database)."""

        if parent is None:
            raise ValueError("'parent' must be provided")

        parent_id = get_parent_id(parent)

        if title is not None:
            properties["title"] = Title.from_value(title)

        props = {name: prop.to_api() for name, prop in properties.items()}

        log.info("Creating page [%s] - %s", parent_id, title)

        data = self().create(parent=parent_id, properties=props, children=children)

        return Page.parse_obj(data)

    # https://developers.notion.com/reference/retrieve-a-page
    def retrieve(self, page_id):
        """Returns the Page with the given ID."""

        log.info("Retrieving page :: %s", page_id)

        return Page.parse_obj(self().retrieve(page_id))

    # https://developers.notion.com/reference/patch-page
    def update(self, page):
        """Updates the Page object on the server.

        The page info will be refreshed to the latest version from the server.
        """

        log.info("Updating page info :: %s", page.id)

        data = self().update(page.id, **page.to_api())

        page.refresh(**data)


class UsersEndpoint(Endpoint):
    """Notional interface to the API 'users' endpoint."""

    def __call__(self):
        return self.session.client.users

    # https://developers.notion.com/reference/get-users
    def list(self):
        """Return an iterator for all users in the workspace."""

        users = EndpointIterator(endpoint=self().list)
        log.info("Listing known users...")

        return ResultSet(session=self.session, src=users, cls=User)

    # https://developers.notion.com/reference/get-user
    def retrieve(self, user_id):
        """Returns the User with the given ID."""

        log.info("Retrieving user :: ", user_id)

        return User.parse_obj(self.users.retrieve(user_id))


class Session(object):
    """An active session with the Notion SDK."""

    def __init__(self, **kwargs):
        self.client = notion_client.Client(**kwargs)

        self.blocks = BlocksEndpoint(self)
        self.databases = DatabasesEndpoint(self)
        self.pages = PagesEndpoint(self)
        # self.search = SearchEndpoint(self) ??
        self.users = UsersEndpoint(self)

        log.info("Initialized Notion SDK client")


def get_parent_id(parent):
    """Return the correct parent ID based on the object type."""

    if isinstance(parent, ParentRef):
        return parent.to_api()
    elif isinstance(parent, Page):
        return {"type": "page_id", "page_id": parent.id.hex}
    elif isinstance(parent, Database):
        return {"type": "database_id", "database_id": parent.id.hex}

    # TODO should we support adding to the workspace?
    raise ValueError("Unrecognized 'parent' attribute")



import os

from setuptools import setup, find_packages
from notional.version import __version__

with open("README.md", "r") as fp:
    readme = fp.read()

requirements = ["notion-client>=0.6", "pydantic>=1.8.2"]

setup(
    name="notional",
    version=__version__,
    author="Jason Heddings",
    author_email="jheddings@gmail.com",
    description="An high-level interface for the Notion SDK.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/jheddings/notional/",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Utilities for working text, markdown & Rich Text in Notion."""

from enum import Enum

# this might be a place to capture other utilities for working with markdown, text
# rich text, etc...  the challenge is not importing types due to a circular ref.


def plain_text(*rtf):
    """Return the combined plain text from the list of RichText objects."""
    return "".join(text.plain_text for text in rtf)


class Color(str, Enum):
    """Basic color values."""

    default = "default"
    gray = "gray"
    brown = "brown"
    orange = "orange"
    yellow = "yellow"
    green = "green"
    blue = "blue"
    purple = "purple"
    pink = "pink"
    red = "red"


class FullColor(str, Enum):
    """Extended color values, including backgounds."""

    default = "default"
    gray = "gray"
    brown = "brown"
    orange = "orange"
    yellow = "yellow"
    green = "green"
    blue = "blue"
    purple = "purple"
    pink = "pink"
    red = "red"

    gray_background = "gray_background"
    brown_background = "brown_background"
    orange_background = "orange_background"
    yellow_background = "yellow_background"
    green_background = "green_background"
    blue_background = "blue_background"
    purple_background = "purple_background"
    pink_background = "pink_background"
    red_background = "red_background"
